﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Ants
{
    public class Colony
    {
        int fNumOfNests;
        //INFO: Real ant speed here
        float fRealSpeed = 20F;
        float fDefaultSpeed = 4.2F;
        float fRealAntSpeedStDev = 6F;
        float fDefaultAntSpeedStDev = 2F;

        World fWorld;
        Dictionary<int, Dictionary<int, List<Pheromone>>> fFoodPheromoneMap = new Dictionary<int, Dictionary<int, List<Pheromone>>>();

        public Dictionary<int, Dictionary<int, List<Pheromone>>> HomePheromoneMap
        {
            get { return fFoodPheromoneMap; }
        }

        public List<Nest> Nests = new List<Nest>();
        public World World
        {
            get { return fWorld; }
        }

        public enum ForagingType
        {
            Individual = 0,
            Group = 1,
            Pheromone = 2
        }

        //Constructor
        public Colony(int aNumOfNests, int aForagers, int aTransporters, World aWorld, string aNestFile)
        {
            fNumOfNests = aNumOfNests;
            fWorld = aWorld;

            int ForagersPerNest = aForagers;
            int TransportersPerNest = aTransporters;

            if (aNestFile == null)
            {
                //INFO: random nests vs fixed....
                //for (int i = 0; i < fNumOfNests; ++i)
                //{
                    //PointF location1 = fWorld.RandomLocation(World.fSize);
                    PointF location1 = new PointF(50, 50);
                    PointF location2 = new PointF(500, 500);
                    PointF location3 = new PointF(500, 50);
                    PointF location4 = new PointF(50, 500);
                    Nests.Add(new Nest(ForagersPerNest, TransportersPerNest, fWorld, location1, this, 0, fDefaultSpeed, fDefaultAntSpeedStDev));
                    Nests.Add(new Nest(ForagersPerNest, TransportersPerNest, fWorld, location2, this, 1, fDefaultSpeed, fDefaultAntSpeedStDev));
                    Nests.Add(new Nest(ForagersPerNest, TransportersPerNest, fWorld, location3, this, 2, fDefaultSpeed, fDefaultAntSpeedStDev));
                    Nests.Add(new Nest(ForagersPerNest, TransportersPerNest, fWorld, location4, this, 3, fDefaultSpeed, fDefaultAntSpeedStDev));
                //}
            }
            else
            {
                int count = 0;
                PointF[] OriginalNestList = World.ReadNestFile(aNestFile, aNumOfNests);
                PointF[] NestList = World.TransformNests(OriginalNestList);
                float fAntSpeed = World.ConvertSpeed(OriginalNestList, fRealSpeed);
                //PointF[] NestList = OriginalNestList;
                foreach (PointF lposition in NestList)
                {
                    Nests.Add(new Nest(ForagersPerNest, TransportersPerNest, fWorld, lposition, this, count, fAntSpeed, fRealAntSpeedStDev));
                    count++;
                }
            }

            foreach (Nest lNest in Nests)
            {
                foreach (Nest locNest in Nests)
                {
                    lNest.ArrivalsFromNests.Add(locNest, 0);
                }
            }
        }

        //Shuffles then steps all nests in colony and updates pheromone map
        public void Step(Graphics graphics)
        {
            fWorld.FYShuffleList(Nests);

            foreach (Nest lNest in Nests)
            {
                lNest.Step(graphics);
            }

            UpdatePheromones();
        }

        //Adds pheromone to map of aType at aPosition
        public void AddPheromone(PointF aPosition, float aAmount, Pheromone.Type aType)
        {
            if ((fWorld.Age % Pheromone.DropInterval) == 0)
            {
                int xIndex = (int)(aPosition.X - (aPosition.X % Pheromone.SensingDistance));
                int yIndex = (int)(aPosition.Y - (aPosition.Y % Pheromone.SensingDistance));
                Dictionary<int, Dictionary<int, List<Pheromone>>> lList;

                switch (aType)
                {
                    case Pheromone.Type.FoundFood:
                        lList = fFoodPheromoneMap;
                        break;

                    default:
                        return;
                }

                if (!lList.ContainsKey(xIndex))
                    lList.Add(xIndex, new Dictionary<int, List<Pheromone>>());
                if (!lList[xIndex].ContainsKey(yIndex))
                    lList[xIndex].Add(yIndex, new List<Pheromone>());

                lList[xIndex][yIndex].Add(new Pheromone(aPosition, aAmount, this.fWorld.Age, aType));
            }
        }

        //Checks for pheromones of aType within aRegion
        public float CheckPheromone(Pheromone.Type aType, RectangleF aRegion)
        {
            int xStartSearch = (int)(aRegion.Left - (aRegion.Left % Pheromone.SensingDistance));
            int yStartSearch = (int)(aRegion.Top - (aRegion.Top % Pheromone.SensingDistance));
            int xEndSearch = (int)(aRegion.Right - (aRegion.Right % Pheromone.SensingDistance));
            int yEndSearch = (int)(aRegion.Bottom - (aRegion.Bottom % Pheromone.SensingDistance));
            Dictionary<int, Dictionary<int, List<Pheromone>>> lList;

            switch (aType)
            {
                case Pheromone.Type.FoundFood:
                    lList = fFoodPheromoneMap;
                    break;

                default:
                    return 0;
            }

            float pheromoneAmount = 0F;

            for (int xIndex = xStartSearch; xIndex <= xEndSearch; xIndex += Pheromone.SensingDistance)
            {
                if (!lList.ContainsKey(xIndex))
                    continue;

                Dictionary<int, List<Pheromone>> xList = lList[xIndex];

                for (int yIndex = yStartSearch; yIndex <= yEndSearch; yIndex += Pheromone.SensingDistance)
                {
                    if (!xList.ContainsKey(yIndex))
                        continue;

                    List<Pheromone> yList = xList[yIndex];

                    for (int i = yList.Count - 1; i >= 0; --i)
                    {
                        Pheromone pheromone = yList[i];

                        if (aRegion.Contains(pheromone.Position))
                            pheromoneAmount += pheromone.GetAmount(World.Age);

                        if (pheromone.Expired(World.Age))
                            yList.RemoveAt(i);
                    }
                }
            }

            return pheromoneAmount;
        }

        //Removes decayed pheromones
        public void UpdatePheromones()
        {
            List<Pheromone> lDelete = new List<Pheromone>();

            foreach (KeyValuePair<int, Dictionary<int, List<Pheromone>>> xPair in this.HomePheromoneMap)
            {
                foreach (KeyValuePair<int, List<Pheromone>> yPair in xPair.Value)
                {
                    foreach (Pheromone pheromone in yPair.Value)
                    {
                        if (pheromone.GetAmount(fWorld.Age) == 0)
                            lDelete.Add(pheromone);
                    }

                    foreach (Pheromone pheromone in lDelete)
                        yPair.Value.Remove(pheromone);

                    lDelete.Clear();
                }
            }
        }

    }
}
